//------------------------------------------------------------------------------
// File: AsyncFlt.cpp
//
// Desc: DirectShow sample code - implementation of CAsyncFilter.
//
// Copyright (c) 1996-2001 Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------------------------


#include <streams.h>
#include <asyncio.h>
#include <asyncrdr.h>
#include <initguid.h>   // Make our GUID get defined

#pragma warning(disable:4710)  // 'function' not inlined (optimization)
#include "asyncflt.h"


//
// Setup data for filter registration
//
const AMOVIESETUP_MEDIATYPE sudOpPinTypes =
{ &MEDIATYPE_Stream     // clsMajorType
, &MEDIASUBTYPE_NULL }; // clsMinorType

const AMOVIESETUP_PIN sudOpPin =
{ L"Output"          // strName
, FALSE              // bRendered
, TRUE               // bOutput
, FALSE              // bZero
, FALSE              // bMany
, &CLSID_NULL        // clsConnectsToFilter
, L"Input"           // strConnectsToPin
, 1                  // nTypes
, &sudOpPinTypes };  // lpTypes

const AMOVIESETUP_FILTER sudAsync =
{ &CLSID_RIFFCDXAReader           // clsID
, L"RIFF/CDXA Source"             // strName
, MERIT_UNLIKELY                  // dwMerit
, 1                               // nPins
, &sudOpPin };                    // lpPin


//
//  Object creation template
//
CFactoryTemplate g_Templates[1] = {
    { L"RIFF/CDXA Source"
    , &CLSID_RIFFCDXAReader
    , CAsyncFilter::CreateInstance
    , NULL
    , &sudAsync }
};

int g_cTemplates = sizeof(g_Templates) / sizeof(g_Templates[0]);

//
// Filter registration functions
//
HRESULT DllRegisterServer()
{
    return AMovieDllRegisterServer2(TRUE);
}

HRESULT DllUnregisterServer()
{
    return AMovieDllRegisterServer2(FALSE);
}


/* Create a new instance of this class */
CUnknown *CAsyncFilter::CreateInstance(LPUNKNOWN pUnk, HRESULT *phr)
{
    /*  DLLEntry does the right thing with the return code and
        returned value on failure
    */
    return new CAsyncFilter(pUnk, phr);
}


/////////////////////////////////////////////////////////////////////////////
//
// stripped file functions
//

LONGLONG actualPosition(LONGLONG strippedPos){
	LONGLONG numFullSectors=strippedPos/SECTOR_USER_DATA;
	LONGLONG lastSector=strippedPos%SECTOR_USER_DATA;
	return RIFF_SIZE+numFullSectors*SECTOR_SIZE+SECTOR_HEADER+lastSector;
}

LONGLONG strippedPosition(LONGLONG actualPosition){
	LONGLONG res=0;
	if (actualPosition>=RIFF_SIZE)
		actualPosition -= RIFF_SIZE;

	LONGLONG lastSector=actualPosition%SECTOR_SIZE;
	LONGLONG numFullSectors=actualPosition/SECTOR_SIZE;
	LONGLONG withoutLast=numFullSectors*SECTOR_USER_DATA;
	dprintf ("lastsector=%d, #(-)sectors=%d, withoutlast=%d",(long)lastSector, (long)numFullSectors,(long)withoutLast);

	if (lastSector < SECTOR_HEADER)
		lastSector=0;
	else if (lastSector >= SECTOR_HEADER + SECTOR_USER_DATA)
		lastSector = SECTOR_USER_DATA;
	else
		lastSector -= SECTOR_HEADER;

	res = withoutLast + lastSector;

	dprintf ("actualPosition=%d, strippledPosition=%d",(long)(actualPosition+RIFF_SIZE), (long)res);

	return res;
}


//////////////////////////////////////////////////////////////////////////////
//
// CMemStream methods
//


CMemStream::CMemStream() :
//        m_llPosition(0)
        m_llPosition(actualPosition(0))
    {
    }

    /*  Initialization */
void CMemStream::Init(HANDLE hFile, LPBYTE pbData, LONGLONG llLength, DWORD dwKBPerSec = INFINITE)
    {
        dprintf ("memstream::init");
		m_pbData = pbData;
        m_llLength = llLength;
		m_strippedLength=strippedPosition(llLength);
        m_dwKBPerSec = dwKBPerSec;
        m_dwTimeStart = timeGetTime();
		m_hFile=hFile;
		dprintf ("llLength=%d, strippedLength=%d",(long)llLength, (long)m_strippedLength);

    }

HRESULT CMemStream::SetPointer(LONGLONG llPos)
    {
		dprintf ("set pointer: %d",llPos);

//        if (llPos < 0 || llPos > m_llLength) {
        if (llPos < 0 || llPos > m_strippedLength) {
			dprintf ("returned: false");
            return S_FALSE;
        } else {
//            m_llPosition = llPos;
            m_strippedPosition = llPos;
            m_llPosition = actualPosition(llPos);
			dprintf ("returned: true");
            return S_OK;
        }
    }

HRESULT CMemStream::Read(PBYTE pbBuffer, DWORD dwBytesToRead, BOOL bAlign, LPDWORD pdwBytesRead)
    {
        CAutoLock lck(&m_csLock);
        DWORD dwReadLength;

        /*  Wait until the bytes are here! */
        DWORD dwTime = timeGetTime();

//        if (m_llPosition + dwBytesToRead > m_llLength) {
        if (m_strippedPosition + dwBytesToRead > m_strippedLength) {
//            dwReadLength = (DWORD)(m_llLength - m_llPosition);
            dwReadLength = (DWORD)(m_strippedLength - m_strippedPosition);
        } else {
            dwReadLength = dwBytesToRead;
        }
        DWORD dwTimeToArrive =
//            ((DWORD)m_llPosition + dwReadLength) / m_dwKBPerSec;
            ((DWORD)m_strippedPosition + dwReadLength) / m_dwKBPerSec;

        if (dwTime - m_dwTimeStart < dwTimeToArrive) {
            Sleep(dwTimeToArrive - dwTime + m_dwTimeStart);
        }

//        CopyMemory((PVOID)pbBuffer, (PVOID)(m_pbData + m_llPosition), dwReadLength);
//        strippedCopyMemory((PVOID)pbBuffer, (PVOID)(m_pbData), m_strippedPosition, dwReadLength);
//        strippedReadFile((PVOID)pbBuffer, (PVOID)(m_pbData), m_strippedPosition, dwReadLength);
        strippedReadFile((PVOID)pbBuffer, m_strippedPosition, dwReadLength);

        m_llPosition += dwReadLength;
		m_strippedPosition+=dwReadLength;
        *pdwBytesRead = dwReadLength;
		dprintf("read: %d bytes @ %d",dwReadLength, m_llPosition-dwReadLength);
        return S_OK;
    }

LONGLONG CMemStream::Size(LONGLONG *pSizeAvailable)
    {
        LONGLONG llCurrentAvailable =
            static_cast <LONGLONG> (UInt32x32To64((timeGetTime() - m_dwTimeStart),m_dwKBPerSec));
 
//       *pSizeAvailable =  min(m_llLength, llCurrentAvailable);
       *pSizeAvailable =  min(m_strippedLength, llCurrentAvailable);
	   dprintf ("size requested: %d",m_strippedLength);
//        return m_llLength;
        return m_strippedLength;
    }

DWORD CMemStream::Alignment()
    {
        return 1;
    }

void CMemStream::Lock()
    {
        m_csLock.Lock();
    }

void CMemStream::Unlock()
    {
        m_csLock.Unlock();
    }

//void CMemStream::strippedReadFile(void* dest, void* src, LONGLONG strippedPos, LONGLONG length)
void CMemStream::strippedReadFile(void* dest, LONGLONG strippedPos, LONGLONG length)
{
	LONGLONG actualStart=RIFF_SIZE + 
						(strippedPos/SECTOR_USER_DATA)*SECTOR_SIZE + 
						SECTOR_HEADER +
						strippedPos%SECTOR_USER_DATA;
	LONGLONG actualStart2=actualPosition(strippedPos);
	if (actualStart!=actualStart2)
		dprintf ("actual pos diff from actualStart");
	
	LONGLONG startForPrintf=actualStart;

	if (length<= (SECTOR_USER_DATA - strippedPos%SECTOR_USER_DATA)){	// inside single USER_DATA block
//		CopyMemory(dest, (char*)(src) + actualStart, (DWORD)length);
		ReadFromFile(dest, actualStart, (DWORD)length);
	} else {	//extends over more than single USER_DATA block
		LONGLONG firstLength=SECTOR_USER_DATA - strippedPos%SECTOR_USER_DATA;
//		CopyMemory(dest, (char*)src+actualStart, (DWORD)firstLength);
		ReadFromFile(dest, actualStart, (DWORD)firstLength);
		length-=firstLength;
		dest=(char*)dest+firstLength;

		actualStart+=firstLength+SECTOR_EDC+SECTOR_HEADER;//points to the start of next USER_DATA block
		while (length>0){
			if (length<=SECTOR_USER_DATA){	// last copy, only part of USER_DATA block
//				CopyMemory(dest, (char*)src+actualStart, (DWORD)length);
				ReadFromFile(dest, actualStart, (DWORD)length);
				break;
			} else {
//				CopyMemory(dest, (char*)src+actualStart, SECTOR_USER_DATA);
				ReadFromFile(dest, actualStart, SECTOR_USER_DATA);
				length-=SECTOR_USER_DATA;
				dest=(char*)dest+SECTOR_USER_DATA;
				actualStart+=SECTOR_SIZE;
			}
		}
	}
	dprintf ("ActualStart(%d)=%d",(long)strippedPos, (long)startForPrintf);

}

//////////////////////////////////////////////////////////////////////////////
//
// CAsyncFilter methods
//

CAsyncFilter::CAsyncFilter(LPUNKNOWN pUnk, HRESULT *phr) :
        CAsyncReader(NAME("Mem Reader"), pUnk, &m_Stream, phr),
			m_pFileName(NULL),
			m_pbData(NULL)
    {
		dprintf ("CasyncFilter::DTOR");
    }

CAsyncFilter::~CAsyncFilter()
    {
        delete [] m_pbData;
        delete [] m_pFileName;
		CloseHandle(m_hFile);
		dprintf ("CasyncFilter::DTOR");
    }

STDMETHODIMP CAsyncFilter::NonDelegatingQueryInterface(REFIID riid, void **ppv)
    {
        if (riid == IID_IFileSourceFilter) {
            return GetInterface((IFileSourceFilter *)this, ppv);
        } else {
            return CAsyncReader::NonDelegatingQueryInterface(riid, ppv);
        }
    }

/*  IFileSourceFilter methods */

//  Load a (new) file
STDMETHODIMP CAsyncFilter::Load(LPCOLESTR lpwszFileName, const AM_MEDIA_TYPE *pmt)
    {
        CheckPointer(lpwszFileName, E_POINTER);

        // lstrlenW is one of the few Unicode functions that works on win95
        int cch = lstrlenW(lpwszFileName) + 1;

#ifndef UNICODE
        TCHAR *lpszFileName=0;
        lpszFileName = new char[cch * 2];
        if (!lpszFileName) {
      	    return E_OUTOFMEMORY;
        }
        WideCharToMultiByte(GetACP(), 0, lpwszFileName, -1,
    			lpszFileName, cch, NULL, NULL);
#else
        TCHAR lpszFileName[MAX_PATH]={0};
        lstrcpy(lpszFileName, lpwszFileName);
#endif
        CAutoLock lck(&m_csFilter);

        /*  Check the file type */
        CMediaType cmt;
        if (NULL == pmt) {
            cmt.SetType(&MEDIATYPE_Stream);
            cmt.SetSubtype(&MEDIASUBTYPE_NULL);
//            cmt.SetSubtype(&MEDIASUBTYPE_Avi);
			dprintf ("pmt=null");
        } else if (pmt->subtype == MEDIASUBTYPE_RIFF_CDXA ) {
            cmt.SetType(&MEDIATYPE_Stream);
            cmt.SetSubtype(&MEDIASUBTYPE_NULL);
//            cmt.SetSubtype(&MEDIASUBTYPE_MPEG1Audio);
//            cmt.SetSubtype(&MEDIASUBTYPE_Avi);
//            cmt.SetSubtype(&MEDIASUBTYPE_Ogg);
			dprintf ("MPEG1Audio");
			dprintf (" pmt->subtype == MEDIASUBTYPE_RIFF_CDXA");
        } else {
			dprintf ("pmt not null and not riff");
            cmt = *pmt;
        }

//        if (!ReadTheFile(lpszFileName)) {
		if (!OpenTheFile(lpszFileName)){
#ifndef UNICODE
            delete [] lpszFileName;
#endif
            return E_FAIL;
        }


		dprintf("calling m_stream.init");
        m_Stream.Init(m_hFile, m_pbData, m_llSize);

///// Detect Media Type //////

		dprintf("detecting Media Type");
		m_pData = new BYTE[SECTOR_USER_DATA];
		m_Stream.ReadFromFile(m_pData, RIFF_SIZE+SECTOR_HEADER, SECTOR_USER_DATA);
		DetectMediaType(m_pData);
		delete [] m_pData;
        
		cmt.SetType(&MediaTypeGUID);
		cmt.SetSubtype(&MediaSubtypeGUID);

//////////////////////////////
		
		m_pFileName = new WCHAR[cch];

        if (m_pFileName!=NULL)
    	    CopyMemory(m_pFileName, lpwszFileName, cch*sizeof(WCHAR));

        // this is not a simple assignment... pointers and format
        // block (if any) are intelligently copied
    	m_mt = cmt;

        /*  Work out file type */
        cmt.bTemporalCompression = TRUE;	       //???
        cmt.lSampleSize = 1;

        return S_OK;
    }

    // Modeled on IPersistFile::Load
    // Caller needs to CoTaskMemFree or equivalent.

STDMETHODIMP CAsyncFilter::GetCurFile(LPOLESTR * ppszFileName, AM_MEDIA_TYPE *pmt)
    {
        CheckPointer(ppszFileName, E_POINTER);
        *ppszFileName = NULL;

        if (m_pFileName!=NULL) {
        	DWORD n = sizeof(WCHAR)*(1+lstrlenW(m_pFileName));

            *ppszFileName = (LPOLESTR) CoTaskMemAlloc( n );
            if (*ppszFileName!=NULL) {
                  CopyMemory(*ppszFileName, m_pFileName, n);
            }
        }

        if (pmt!=NULL) {
            CopyMediaType(pmt, &m_mt);
        }

        return NOERROR;
    }


/////////////////////////////////////////////////////////////////////////////
//
// File reading stuff
//

//BOOL CAsyncFilter::ReadTheFile(LPCTSTR lpszFileName)
BOOL CAsyncFilter::OpenTheFile(LPCTSTR lpszFileName)
{
    /*  Read the file */
//    HANDLE hFile = CreateFile(lpszFileName,
    m_hFile = CreateFile(lpszFileName,
                              GENERIC_READ,
                              FILE_SHARE_READ,
                              NULL,
                              OPEN_EXISTING,
                              0,
                              NULL);
    if (m_hFile == INVALID_HANDLE_VALUE) {
        DbgLog((LOG_TRACE, 2, TEXT("Could not open %s\n"), lpszFileName));
        return FALSE;
    }

    ULARGE_INTEGER uliSize;
    uliSize.LowPart = GetFileSize(m_hFile, &uliSize.HighPart);

/*
    PBYTE pbMem = new BYTE[uliSize.LowPart];
    if (pbMem == NULL) {
        CloseHandle(hFile);
        return FALSE;
    }

    DWORD dwBytesRead;
    if (!ReadFile(hFile,
                  (LPVOID)pbMem,
                  uliSize.LowPart,
                  &dwBytesRead,
                  NULL) ||
        dwBytesRead != uliSize.LowPart) {
        DbgLog((LOG_TRACE, 1, TEXT("Could not read file\n")));
        delete [] pbMem;
        CloseHandle(hFile);
        return FALSE;
    }
*/
	PBYTE pbMem = NULL;//new
    m_pbData = pbMem;
    m_llSize = (LONGLONG)uliSize.QuadPart;
//    CloseHandle(hFile);

    return TRUE;
}


BOOL CMemStream::ReadFromFile(void *dest, LONGLONG pos, DWORD size)
{

    DWORD dwBytesRead;
    LARGE_INTEGER liPos;

	liPos.QuadPart = pos;
		
	SetFilePointer(m_hFile, liPos.LowPart, &liPos.HighPart , FILE_BEGIN);
	
	if (!ReadFile(m_hFile,
                  (LPVOID)dest,
                  size,
                  &dwBytesRead,
                  NULL) ||
        dwBytesRead != size) {
        DbgLog((LOG_TRACE, 1, TEXT("Could not read file\n")));
//        CloseHandle(hFile);
        return FALSE;
    }
	
	return TRUE;
}



int CopyToken (char *dest, char *szValue, int *i)
{
	
	int d = 0;
	while(*i < (int)strlen(szValue))
	{
		if (szValue[*i] == ' ') { *i += 1; continue; }
		if (szValue[*i] == ',') { *i += 1; break; }
		dest[d++] = szValue[*i];
		*i += 1;
	}
	dest[d] = '\0';

	if (*i == (int)strlen(szValue)) return 1; // end of string

	return 0;
}


BOOL CAsyncFilter::DetectMediaType(BYTE *Data)
{

	#define BUFSIZE MAX_PATH+1

	HKEY hkHKCR_MediaTypeRoot;
	HKEY hkMediaType;
	HKEY hkMediaSubtype;
	CHAR szValueIndex[2];
	TCHAR szValue[BUFSIZE];
	DWORD dwValueSize;

	CHAR szMediaSubtypeName[MAX_PATH] = {0};

	BOOL Found = FALSE;

	
	LPOLESTR wszMediaType;
	CHAR szMediaTypeName[MAX_PATH];
	StringFromCLSID(MEDIATYPE_Stream, &wszMediaType);
	if (WideCharToMultiByte(CP_ACP, 0, wszMediaType, -1, szMediaTypeName, MAX_PATH, NULL, NULL) == 0)
	CoTaskMemFree(&wszMediaType);

	dprintf("MediaType Key: %s\n",szMediaTypeName);

	RegOpenKeyEx(HKEY_CLASSES_ROOT, "Media Type", NULL, KEY_ALL_ACCESS, &hkHKCR_MediaTypeRoot);
	RegOpenKeyEx(hkHKCR_MediaTypeRoot, szMediaTypeName, NULL, KEY_ALL_ACCESS, &hkMediaType);

	dprintf("MediaSubtype Keys:\n");

	int i = 0;
	while(!Found && RegEnumKey(hkMediaType, i++, szMediaSubtypeName, sizeof(szMediaSubtypeName)) == ERROR_SUCCESS)
	{
		RegOpenKeyEx(hkMediaType, szMediaSubtypeName, NULL, KEY_ALL_ACCESS, &hkMediaSubtype);
		
		dprintf ("%s\n",szMediaSubtypeName);

		int j = 0;
		while (!Found && j < 10 && (dwValueSize = BUFSIZE, RegQueryValueEx(hkMediaSubtype, _itoa(j++, szValueIndex, 10), NULL, NULL, (LPBYTE)szValue, &dwValueSize) == ERROR_SUCCESS))
		{
			// parse szValue and compare with Data
			// once found, set Found = TRUE

			dprintf("\"%s\"=\"%s\"\n",szValueIndex,szValue);
			
			LONG offset, length;
			char mask[MAX_PATH], content[MAX_PATH], tmpstr[MAX_PATH];

			int m = 0;
			int is_end = 0;
			int failed = 0;

			while(!is_end && !failed)
			{

				CopyToken(tmpstr, szValue, &m);
				offset = atoi(tmpstr);
				CopyToken(tmpstr, szValue, &m);
				length = atoi(tmpstr);
				CopyToken(mask, szValue, &m);
				is_end = CopyToken(content, szValue, &m);

//				dprintf("data: %d,%d,\"%s\",\"%s\"", offset, length, mask, content);

				int l = 0;

//				for (l = 0; l < length; l++)
//					dprintf("%x",Data[l+offset]);

				char datahexstr[3];
				for (l = 0; l < length && !failed; l++)
				{
					_itoa(Data[l+offset], datahexstr, 16); 

					if (_strnicmp(datahexstr, &(content[l*2]), 2) != 0)  // TODO: implement mask
						failed = 1;
					else 
						dprintf("%s (%c)", datahexstr,Data[l+offset]);

				}
			}

			if (!failed) Found = TRUE; // FOUND!!!

		}

		RegCloseKey(hkMediaSubtype);
	}

	RegCloseKey(hkMediaType);
	RegCloseKey(hkHKCR_MediaTypeRoot);
	
	if (Found)
	{
		MediaTypeGUID = MEDIATYPE_Stream;

		WCHAR wszSubtypeName[MAX_PATH];
		MultiByteToWideChar(CP_ACP, 0, szMediaSubtypeName, -1, wszSubtypeName, MAX_PATH);
		CLSIDFromString(wszSubtypeName, &MediaSubtypeGUID);
	}
	else
	{
		MediaTypeGUID = MEDIATYPE_Stream;
		MediaSubtypeGUID = MEDIASUBTYPE_NULL;
	}
	
	if (Found)
		dprintf("Found!");
	else
		dprintf("Not found...");

	return Found;
}

